"""Lineage payload builders for session analysis."""

from __future__ import annotations

from collections import Counter
from typing import Any

from gotta.content.model import ContentSnapshot
from gotta.content.path import content_locator
from gotta.content.store import scan_content_store
from gotta.lead.aggregate import aggregate_lead_sources
from gotta.lead.edge import build_lead_edge_records
from gotta.lead.snapshot import (
    snapshot_artifact_locator,
    snapshot_display_name,
    snapshot_locator,
)
from gotta.source.visibility import best_visibility_metadata

from ..core import (
    artifact_kind,
    follow_command,
    lead_kind,
    provider_name,
    rendered_actor,
    render_variant,
    render_variant_label,
    resolved_visibility_metadata,
    resource_label,
    session_read_command,
    topology_next_step,
)
from ..manifest.record import manifest_entries
from .focus import focus_match_threshold, ordered_focus_scan_entries


def _revision_edges(snapshots: list[ContentSnapshot]) -> list[dict[str, str]]:
    tracks: dict[tuple[str, tuple[str, str]], list[dict[str, str]]] = {}
    for snapshot in snapshots:
        canonical = str(
            snapshot.metadata.get("canonical_locator", "")
            or snapshot.metadata.get("locator", "")
        ).strip()
        if not canonical:
            continue
        variant = render_variant(snapshot)
        for event in snapshot.events:
            tracks.setdefault((canonical, variant), []).append(
                {
                    "timestamp": event.timestamp,
                    "digest": snapshot.digest,
                    "preferred_name": str(
                        snapshot.metadata.get("preferred_name", "") or event.link_name
                    ),
                    "plugin": str(snapshot.metadata.get("plugin", "") or "unknown"),
                    "actor": rendered_actor(
                        snapshot.metadata.get("actor"),
                        session_root=snapshot.content_dir.parent.parent,
                    ),
                    "rendering": render_variant_label(variant),
                }
            )
    edges: list[dict[str, str]] = []
    for (locator, _variant), items in sorted(tracks.items()):
        prior_item: dict[str, str] | None = None
        for item in sorted(
            items, key=lambda current: (current["timestamp"], current["digest"])
        ):
            if prior_item is None:
                prior_item = item
                continue
            if item["digest"] == prior_item["digest"]:
                prior_item = item
                continue
            edges.append(
                {
                    "locator": locator,
                    "preferredName": item["preferred_name"]
                    or prior_item["preferred_name"],
                    "from": prior_item["digest"],
                    "to": item["digest"],
                    "fromTimestamp": prior_item["timestamp"],
                    "toTimestamp": item["timestamp"],
                    "plugin": item["plugin"] or prior_item["plugin"],
                    "actor": item["actor"] or prior_item["actor"],
                    "rendering": item["rendering"] or prior_item["rendering"],
                }
            )
            prior_item = item
    return edges


def lineage_payload(dirs, *, session_ref: str = "") -> dict[str, Any]:
    snapshots = scan_content_store(dirs.content_dir)
    snapshot_by_digest = {snapshot.digest: snapshot for snapshot in snapshots}
    entries: list[dict[str, Any]] = [dict(entry) for entry in manifest_entries(dirs)]
    source_map: dict[str, dict[str, Any]] = {}
    edge_plugins: dict[tuple[str, str], list[str]] = {}
    edge_actors: dict[tuple[str, str], set[str]] = {}
    content_details: dict[str, dict[str, set[str]]] = {}

    for entry in entries:
        source = str(
            entry.get("canonical_locator") or entry.get("locator") or "unknown"
        )
        checksum = str(entry.get("checksum") or "")
        if not checksum:
            continue
        source_state = source_map.setdefault(
            source,
            {
                "content": set(),
                "locators": set(),
                "plugins": set(),
                "actors": set(),
                "artifact_kinds": set(),
                "entries": 0,
                "variants": set(),
                "visibility": {},
            },
        )
        source_state["content"].add(checksum)
        locator = str(entry.get("locator") or source)
        source_state["locators"].add(locator)
        plugin = str(entry.get("plugin") or "unknown")
        actor = rendered_actor(entry.get("actor"), session_root=dirs.session_dir)
        source_state["plugins"].add(plugin)
        source_state["actors"].add(actor)
        kind = artifact_kind(entry.get("artifact_kind"))
        if kind:
            source_state["artifact_kinds"].add(kind)
        source_state["entries"] = int(source_state["entries"]) + 1
        source_state["visibility"] = best_visibility_metadata(
            source_state.get("visibility", {}),
            resolved_visibility_metadata(
                entry,
                provider=str(plugin),
                plugin=str(plugin),
                subcommand=str(entry.get("subcommand") or ""),
                locator=str(source),
            ),
        )
        snapshot = snapshot_by_digest.get(str(checksum))
        if snapshot is not None:
            source_state["variants"].add(render_variant(snapshot))
        edge_plugins.setdefault((source, checksum), []).append(plugin)
        edge_actors.setdefault((source, checksum), set()).add(actor)
        detail = content_details.setdefault(
            checksum,
            {
                "providers": set(),
                "actors": set(),
                "resource_hints": set(),
            },
        )
        detail["providers"].add(
            provider_name(source, plugins=[plugin], fallback=plugin)
        )
        detail["actors"].add(actor)
        source_kind, source_label = resource_label(source)
        if source_kind and source_label:
            detail["resource_hints"].add(f"{source_kind}:{source_label}")
        else:
            detail["resource_hints"].add(source)

    name_counts = Counter(snapshot_display_name(snapshot) for snapshot in snapshots)

    content = [
        {
            "checksum": snapshot.digest,
            "preferredName": snapshot_display_name(snapshot),
            "artifactKind": artifact_kind(snapshot.metadata.get("artifact_kind")),
            "contentLocator": content_locator(snapshot.digest),
            "artifactLocator": snapshot_artifact_locator(snapshot),
            "followCommand": session_read_command(
                snapshot_artifact_locator(snapshot),
                session_ref=session_ref,
            ),
            "nameCollision": name_counts[snapshot_display_name(snapshot)] > 1,
            "nameCount": len(snapshot.names),
            "fetchCount": len(snapshot.events),
            "names": snapshot.names,
            "firstFetchedAt": snapshot.events[0].timestamp if snapshot.events else "",
            "lastFetchedAt": snapshot.events[-1].timestamp if snapshot.events else "",
            "providers": sorted(
                content_details.get(snapshot.digest, {}).get("providers", set())
            ),
            "actors": sorted(
                content_details.get(snapshot.digest, {}).get("actors", set())
            ),
            "resourceHints": sorted(
                content_details.get(snapshot.digest, {}).get("resource_hints", set())
            ),
            **resolved_visibility_metadata(
                dict(snapshot.metadata),
                provider=str(snapshot.metadata.get("plugin") or ""),
                plugin=str(snapshot.metadata.get("plugin") or ""),
                subcommand=str(snapshot.metadata.get("subcommand") or ""),
                locator=str(snapshot_locator(snapshot)),
            ),
        }
        for snapshot in snapshots
    ]
    sources = [
        {
            "locator": locator,
            "contentCount": len(state["content"]),
            "entryCount": int(state["entries"]),
            "artifactKind": (
                next(iter(state["artifact_kinds"]))
                if len(state["artifact_kinds"]) == 1
                else ""
            ),
            "artifactKinds": sorted(str(value) for value in state["artifact_kinds"]),
            "plugins": sorted(str(value) for value in state["plugins"]),
            "actors": sorted(str(value) for value in state["actors"]),
            "locators": sorted(str(value) for value in state["locators"]),
            "collision": False,
            "duplicateMaterialization": len(state["content"]) > 1
            and len(state["variants"]) <= 1,
            "variant": len(state["variants"]) > 1,
            "variantCount": len(state["variants"]),
            "variants": [
                render_variant_label(variant) for variant in sorted(state["variants"])
            ],
            **best_visibility_metadata(state.get("visibility", {})),
        }
        for locator, state in sorted(source_map.items())
    ]
    source_edges = [
        {
            "source": source,
            "checksum": checksum,
            "plugins": sorted(plugins),
            "actors": sorted(edge_actors.get((source, checksum), set())),
            "count": len(plugins),
        }
        for (source, checksum), plugins in sorted(edge_plugins.items())
    ]
    revision_edges = _revision_edges(snapshots)
    lead_edges = build_lead_edge_records(
        snapshots,
        entries,
        classify_kind=lead_kind,
    )
    lead_sources = aggregate_lead_sources(lead_edges)
    for lead in lead_sources:
        locator = str(lead.get("locator") or "").strip()
        if locator:
            lead["followCommand"] = follow_command(locator, session_ref=session_ref)
    collisions = [source["locator"] for source in sources if source["collision"]]
    duplicate_materializations = [
        source["locator"]
        for source in sources
        if source.get("duplicateMaterialization")
    ]
    variants = [source["locator"] for source in sources if source.get("variant")]
    name_collisions = sorted(
        name for name, count in name_counts.items() if count > 1 and name
    )
    materialized_lead_count = sum(
        1 for source in lead_sources if bool(source["materialized"])
    )
    empty = (
        not sources
        and not content
        and not source_edges
        and not revision_edges
        and not lead_edges
    )
    discovery_count = sum(
        1 for item in content if item.get("artifactKind") == "discovery"
    )
    evidence_count = sum(
        1 for item in content if item.get("artifactKind") == "evidence"
    )
    return {
        "sessionDir": str(dirs.session_dir),
        "contentDir": str(dirs.content_dir),
        "manifestPath": str(dirs.content_dir / "manifest.jsonl"),
        "manifestEntryCount": len(entries),
        "contentCount": len(content),
        "sourceCount": len(sources),
        "sourceEdgeCount": len(source_edges),
        "revisionEdgeCount": len(revision_edges),
        "discoveryArtifactCount": discovery_count,
        "evidenceArtifactCount": evidence_count,
        "collisionCount": len(collisions),
        "collisions": collisions,
        "duplicateMaterializationCount": len(duplicate_materializations),
        "duplicateMaterializations": duplicate_materializations,
        "variantCount": len(variants),
        "variants": variants,
        "nameCollisionCount": len(name_collisions),
        "nameCollisions": name_collisions,
        "leadSourceCount": len(lead_sources),
        "materializedLeadSourceCount": materialized_lead_count,
        "unmaterializedLeadSourceCount": len(lead_sources) - materialized_lead_count,
        "leadEdgeCount": len(lead_edges),
        "empty": empty,
        "nextStep": topology_next_step(
            discovery_count=discovery_count,
            evidence_count=evidence_count,
        ),
        "sources": sources,
        "content": content,
        "sourceEdges": source_edges,
        "revisionEdges": revision_edges,
        "leadSources": lead_sources,
        "leadEdges": lead_edges,
    }


def _lineage_source_candidate(item: dict[str, Any]) -> dict[str, Any]:
    return {
        "kind": "source",
        "label": str(item.get("locator") or ""),
        "locator": str(item.get("locator") or ""),
        "artifactKind": str(item.get("artifactKind") or ""),
        "materialized": True,
        "followCommand": str(item.get("followCommand") or ""),
    }


def _lineage_content_candidate(item: dict[str, Any]) -> dict[str, Any]:
    return {
        "kind": "content",
        "label": str(item.get("preferredName") or ""),
        "checksum": str(item.get("checksum") or ""),
        "artifactLocator": str(item.get("artifactLocator") or ""),
        "contentLocator": str(item.get("contentLocator") or ""),
        "artifactKind": str(item.get("artifactKind") or ""),
        "materialized": True,
        "followCommand": str(item.get("followCommand") or ""),
    }


def _lineage_lead_candidate(item: dict[str, Any]) -> dict[str, Any]:
    return {
        "kind": "lead",
        "label": str(item.get("locator") or ""),
        "locator": str(item.get("locator") or ""),
        "artifactKind": str(item.get("artifactKind") or ""),
        "materialized": bool(item.get("materialized")),
        "followCommand": str(item.get("followCommand") or ""),
    }


def _lineage_focus_score(item: dict[str, Any], query: str) -> tuple[int, int, str]:
    query_lower = query.lower()
    candidates = [
        str(item.get("label") or ""),
        str(item.get("locator") or ""),
        str(item.get("artifactLocator") or ""),
        str(item.get("contentLocator") or ""),
        str(item.get("checksum") or ""),
    ]
    lowered = [candidate.lower() for candidate in candidates if candidate]
    score = 0
    if any(candidate == query_lower for candidate in lowered):
        score = 5
    elif any(candidate.startswith(query_lower) for candidate in lowered):
        score = 4
    elif any(f":{query_lower}" in candidate for candidate in lowered):
        score = 3
    elif any(query_lower in candidate for candidate in lowered):
        score = 2
    materialized = 1 if bool(item.get("materialized")) else 0
    label = str(item.get("label") or item.get("locator") or "")
    return (score, materialized, label.lower())


def _empty_lineage_focus_payload(
    payload: dict[str, Any],
    *,
    query: str,
    next_step: str,
) -> dict[str, Any]:
    return {
        "sessionDir": payload["sessionDir"],
        "contentDir": payload["contentDir"],
        "focus": query,
        "matched": False,
        "empty": True,
        "nextStep": next_step,
        "root": {},
        "neighbors": [],
        "sources": [],
        "content": [],
        "sourceEdges": [],
        "revisionEdges": [],
        "leadSources": [],
        "leadEdges": [],
        "discoveryArtifactCount": 0,
        "evidenceArtifactCount": 0,
        "anchors": [],
        "matchedCount": 0,
    }


def lineage_focus_payload(
    payload: dict[str, Any],
    *,
    focus: str,
    limit: int,
    scan_payload: dict[str, Any] | None = None,
) -> dict[str, Any]:
    query = focus.strip()
    if not query:
        return _empty_lineage_focus_payload(
            payload,
            query="",
            next_step="Provide a focus keyword, locator, artifact name, or checksum prefix.",
        )
    sources = [dict(item) for item in payload.get("sources") or []]
    content_items = [dict(item) for item in payload.get("content") or []]
    lead_sources = [dict(item) for item in payload.get("leadSources") or []]
    source_index = {str(item.get("locator") or ""): item for item in sources}
    content_index = {str(item.get("checksum") or ""): item for item in content_items}
    lead_index = {str(item.get("locator") or ""): item for item in lead_sources}
    candidates = [
        *(_lineage_source_candidate(item) for item in sources),
        *(_lineage_content_candidate(item) for item in content_items),
        *(
            _lineage_lead_candidate(item)
            for item in lead_sources
            if str(item.get("locator") or "") not in source_index
        ),
    ]
    matches = sorted(
        (
            candidate
            for candidate in candidates
            if _lineage_focus_score(candidate, query)[0] > 0
        ),
        key=lambda candidate: _lineage_focus_score(candidate, query),
        reverse=True,
    )
    scan_entries = ordered_focus_scan_entries(
        scan_payload,
        limit=max(limit * 2, 8),
    )
    seed_cap = max(4, min(max(limit, 1), 12))
    no_match_step = (
        f"No analyzed lineage anchor or projected artifact matched `{query}`. Try a canonical locator, "
        "artifact name, checksum prefix, or a tighter target from session scan, leads, or manifest."
    )
    if not matches and not scan_entries:
        return _empty_lineage_focus_payload(
            payload,
            query=query,
            next_step=no_match_step,
        )
    best_score = _lineage_focus_score(matches[0], query)[0] if matches else 0
    threshold = focus_match_threshold(best_score)
    seeds: list[dict[str, Any]] = []
    seen_seed_keys: set[tuple[str, str]] = set()

    def add_seed(candidate: dict[str, Any]) -> None:
        kind = str(candidate.get("kind") or "")
        if kind == "source":
            key = ("source", str(candidate.get("locator") or ""))
        elif kind == "content":
            key = ("content", str(candidate.get("checksum") or ""))
        else:
            key = ("lead", str(candidate.get("locator") or ""))
        if not key[1] or key in seen_seed_keys:
            return
        seen_seed_keys.add(key)
        seeds.append(candidate)

    for candidate in matches:
        if _lineage_focus_score(candidate, query)[0] < threshold:
            break
        add_seed(dict(candidate))
        if len(seeds) >= seed_cap:
            break
    for entry in scan_entries:
        checksum = str(entry.get("checksum") or "").strip()
        locator = str(
            entry.get("canonical_locator") or entry.get("locator") or ""
        ).strip()
        if checksum and checksum in content_index:
            add_seed(_lineage_content_candidate(content_index[checksum]))
        if locator and locator in source_index:
            add_seed(_lineage_source_candidate(source_index[locator]))
        if len(seeds) >= seed_cap:
            break

    if not seeds:
        return _empty_lineage_focus_payload(
            payload,
            query=query,
            next_step=no_match_step,
        )

    root = dict(seeds[0])
    selected_sources: set[str] = set()
    selected_content: set[str] = set()
    selected_leads: set[str] = set()
    matched_sources = {
        str(candidate.get("locator") or "")
        for candidate in matches
        if str(candidate.get("kind") or "") == "source"
    }
    matched_leads = {
        str(candidate.get("locator") or "")
        for candidate in matches
        if str(candidate.get("kind") or "") == "lead"
    }
    for candidate in seeds:
        if candidate["kind"] == "source":
            selected_sources.add(str(candidate.get("locator") or ""))
        elif candidate["kind"] == "content":
            selected_content.add(str(candidate.get("checksum") or ""))
        else:
            selected_leads.add(str(candidate.get("locator") or ""))

    def expand_source_and_revision_edges() -> None:
        for edge in payload.get("sourceEdges") or []:
            source = str(edge.get("source") or "")
            checksum = str(edge.get("checksum") or "")
            if source in selected_sources:
                selected_content.add(checksum)
            if checksum in selected_content:
                selected_sources.add(source)
        for edge in payload.get("revisionEdges") or []:
            from_checksum = str(edge.get("from") or "")
            to_checksum = str(edge.get("to") or "")
            if from_checksum in selected_content:
                selected_content.add(to_checksum)
            if to_checksum in selected_content:
                selected_content.add(from_checksum)

    expand_source_and_revision_edges()
    for edge in payload.get("leadEdges") or []:
        source_checksum = str(edge.get("sourceChecksum") or "")
        target_locator = str(edge.get("targetLocator") or "")
        target_is_source = target_locator in source_index
        target_matches_focus = (
            target_locator in matched_sources or target_locator in matched_leads
        )
        if source_checksum in selected_content and (
            target_locator in selected_sources
            or target_locator in selected_leads
            or target_matches_focus
        ):
            if target_is_source:
                selected_sources.add(target_locator)
            else:
                selected_leads.add(target_locator)
        if target_locator in selected_sources or target_locator in selected_leads:
            selected_content.add(source_checksum)
    expand_source_and_revision_edges()

    neighbor_candidates: list[dict[str, Any]] = []
    for locator in sorted(selected_sources):
        if ("source", locator) in seen_seed_keys:
            continue
        source_item = source_index.get(locator)
        if source_item is None:
            continue
        neighbor_candidates.append(
            {
                "kind": "source",
                "label": locator,
                "relation": "materialized source",
                "followCommand": str(source_item.get("followCommand") or ""),
                "artifactKind": str(source_item.get("artifactKind") or ""),
                "materialized": True,
            }
        )
    for checksum in sorted(selected_content):
        if ("content", checksum) in seen_seed_keys:
            continue
        content_item = content_index.get(checksum)
        if content_item is None:
            continue
        neighbor_candidates.append(
            {
                "kind": "content",
                "label": str(content_item.get("preferredName") or checksum),
                "relation": "stored artifact",
                "followCommand": str(content_item.get("followCommand") or ""),
                "artifactKind": str(content_item.get("artifactKind") or ""),
                "materialized": True,
            }
        )
    for locator in sorted(selected_leads):
        if ("lead", locator) in seen_seed_keys:
            continue
        lead_item = lead_index.get(locator)
        if lead_item is None:
            continue
        neighbor_candidates.append(
            {
                "kind": "lead",
                "label": locator,
                "relation": "followable lead",
                "followCommand": str(lead_item.get("followCommand") or ""),
                "artifactKind": str(lead_item.get("artifactKind") or ""),
                "materialized": bool(lead_item.get("materialized")),
            }
        )
    ordered_neighbors = sorted(
        neighbor_candidates,
        key=lambda item: (
            1 if bool(item.get("materialized")) else 0,
            1 if str(item.get("kind") or "") == "content" else 0,
            str(item.get("label") or "").lower(),
        ),
        reverse=True,
    )[: max(limit, 0)]

    neighbor_source_labels = {
        str(item.get("label") or "")
        for item in ordered_neighbors
        if str(item.get("kind") or "") == "source"
    }
    neighbor_content_labels = {
        str(item.get("label") or "")
        for item in ordered_neighbors
        if str(item.get("kind") or "") == "content"
    }
    neighbor_content_checksums = {
        checksum
        for checksum, item in content_index.items()
        if str(item.get("preferredName") or checksum) in neighbor_content_labels
    }
    neighbor_lead_labels = {
        str(item.get("label") or "")
        for item in ordered_neighbors
        if str(item.get("kind") or "") == "lead"
    }
    seed_source_labels = {value for kind, value in seen_seed_keys if kind == "source"}
    seed_content_checksums = {
        value for kind, value in seen_seed_keys if kind == "content"
    }
    seed_lead_labels = {value for kind, value in seen_seed_keys if kind == "lead"}
    selected_sources = {
        locator
        for locator in selected_sources
        if locator in seed_source_labels or locator in neighbor_source_labels
    }
    selected_content = {
        checksum
        for checksum in selected_content
        if checksum in seed_content_checksums or checksum in neighbor_content_checksums
    }
    selected_leads = {
        locator
        for locator in selected_leads
        if locator in seed_lead_labels or locator in neighbor_lead_labels
    }

    selected_source_items = [
        item for item in sources if str(item.get("locator") or "") in selected_sources
    ]
    selected_content_items = [
        item
        for item in content_items
        if str(item.get("checksum") or "") in selected_content
    ]
    selected_lead_items = [
        item
        for item in lead_sources
        if str(item.get("locator") or "") in selected_leads
        and str(item.get("locator") or "") not in selected_sources
    ]
    selected_source_edges = [
        edge
        for edge in payload.get("sourceEdges") or []
        if str(edge.get("source") or "") in selected_sources
        and str(edge.get("checksum") or "") in selected_content
    ]
    selected_revision_edges = [
        edge
        for edge in payload.get("revisionEdges") or []
        if str(edge.get("from") or "") in selected_content
        and str(edge.get("to") or "") in selected_content
    ]
    selected_lead_edges = [
        edge
        for edge in payload.get("leadEdges") or []
        if str(edge.get("sourceChecksum") or "") in selected_content
        and str(edge.get("targetLocator") or "")
        in selected_leads.union(selected_sources)
    ]
    discovery_count = sum(
        1
        for item in selected_content_items
        if str(item.get("artifactKind") or "") == "discovery"
    )
    evidence_count = sum(
        1
        for item in selected_content_items
        if str(item.get("artifactKind") or "") == "evidence"
    )
    return {
        "sessionDir": payload["sessionDir"],
        "contentDir": payload["contentDir"],
        "manifestPath": payload["manifestPath"],
        "focus": query,
        "matched": True,
        "empty": False,
        "nextStep": "",
        "root": root,
        "anchors": seeds[1:],
        "matchedCount": len(seeds),
        "neighbors": ordered_neighbors,
        "sources": selected_source_items,
        "content": selected_content_items,
        "sourceEdges": selected_source_edges,
        "revisionEdges": selected_revision_edges,
        "leadSources": selected_lead_items,
        "leadEdges": selected_lead_edges,
        "discoveryArtifactCount": discovery_count,
        "evidenceArtifactCount": evidence_count,
        "contentCount": len(selected_content_items),
        "sourceCount": len(selected_source_items),
        "sourceEdgeCount": len(selected_source_edges),
        "revisionEdgeCount": len(selected_revision_edges),
        "leadSourceCount": len(selected_lead_items),
        "leadEdgeCount": len(selected_lead_edges),
        "collisionCount": 0,
        "collisions": [],
        "duplicateMaterializationCount": 0,
        "duplicateMaterializations": [],
        "variantCount": 0,
        "variants": [],
    }



"""Argument parsing for `gotta notes`."""

from __future__ import annotations

import argparse

from gotta.session import charter as session_charter


def _add_root_args(parser: argparse.ArgumentParser) -> None:
    session_charter.add_target_args(parser)


def build_parser(command_name: str = "gotta notes") -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=command_name,
        description=(
            "Inspect or mutate canonical actor notes inside the active session "
            "surface. Notes are the canonical actor-authored narration surface, "
            "and short one-line notes are valid."
        ),
    )
    parser.add_argument(
        "action",
        nargs="?",
        choices=["show", "append"],
        help="show one or all actor note surfaces, or append a actor note",
    )
    parser.add_argument("value", nargs="*")
    _add_root_args(parser)
    parser.add_argument(
        "--from-file",
        help="read actor note text from a UTF-8 file instead of inline text; use '-' for stdin",
    )
    parser.add_argument(
        "--stdin",
        dest="use_stdin",
        action="store_true",
        help="read actor note text from stdin explicitly",
    )
    parser.add_argument("--output", choices=["json", "text"], default="text")
    return parser


def session_access_mode(argv: list[str]) -> str:
    positionals = session_charter.argv_positionals(
        argv,
        valued_flags=(
            "--session",
            "--actor",
            "--from-file",
            "--output",
        ),
    )
    action = positionals[0] if positionals else "show"
    return "write" if action == "append" else "read"


def normalize_argv(argv: list[str]) -> list[str]:
    if not argv:
        return argv
    first = argv[0]
    if first in {"show", "append"} or first.startswith("-"):
        return argv
    return ["show", *argv]



"""Rank lead edges and aggregated lead sources."""

from __future__ import annotations

from .canon import low_signal_url_penalty

FIRST_PARTY_LEAD_PROVIDERS = {
    "jira",
    "confluence",
    "slack",
    "github",
    "gdocs",
    "gdrive",
    "gsheets",
    "artifact",
    "content",
}
RELATION_PRIORITY = {
    "mentions": 3,
    "links_to": 2,
    "suggests_search": 1,
}


def is_first_party_target(*, provider: str, kind: str) -> bool:
    return provider in FIRST_PARTY_LEAD_PROVIDERS or kind != "url"


def edge_best_first_sort_key(item: dict[str, object]) -> tuple[object, ...]:
    raw_examples = [str(value) for value in item.get("rawExamples") or [] if str(value)]
    query = raw_examples[0] if raw_examples else str(item.get("targetLocator") or "")
    source_rank = int(item.get("sourceRank") or 0)
    return (
        not bool(item.get("firstParty")),
        not bool(item.get("materialized")),
        _lead_signal_penalty(item),
        bool(item.get("searchSeed")),
        bool(item.get("sourceSearchLike")),
        -_relation_priority(str(item.get("relation") or "")),
        -int(item.get("occurrenceCount") or 0),
        source_rank
        if bool(item.get("sourceSearchLike")) and source_rank > 0
        else 1_000_000,
        *_search_seed_signal_sort_key(query),
        str(item.get("targetLocator") or "").casefold(),
    )


def lead_source_best_first_sort_key(item: dict[str, object]) -> tuple[object, ...]:
    artifact_count = int(item.get("artifactCount") or 0)
    search_like_source_count = int(item.get("searchLikeSourceCount") or 0)
    query = str(item.get("exampleRaw") or item.get("locator") or "")
    best_search_rank = int(item.get("bestSearchRank") or 0)
    return (
        not bool(item.get("firstParty")),
        not bool(item.get("materialized")),
        _lead_signal_penalty(item),
        bool(item.get("searchSeed")),
        bool(artifact_count and search_like_source_count >= artifact_count),
        bool(search_like_source_count),
        -_relation_set_priority(
            [str(value) for value in item.get("relationKinds") or [] if str(value)]
        ),
        -artifact_count,
        -int(item.get("occurrenceCount") or 0),
        best_search_rank if best_search_rank > 0 else 1_000_000,
        *_search_seed_signal_sort_key(query),
        str(item.get("locator") or "").casefold(),
    )


def _relation_priority(value: str) -> int:
    return RELATION_PRIORITY.get(value, 0)


def _relation_set_priority(values: set[str] | list[str]) -> int:
    return max((_relation_priority(str(value)) for value in values), default=0)


def _search_seed_signal_sort_key(query: str) -> tuple[object, ...]:
    tokens = [token for token in query.split() if token]
    uppercase_count = sum(1 for token in tokens if token.isupper())
    titlecase_count = sum(
        1 for token in tokens if token[:1].isupper() and not token.isupper()
    )
    long_lowercase_count = sum(
        1 for token in tokens if not token[:1].isupper() and len(token.casefold()) >= 8
    )
    return (
        -uppercase_count,
        -(titlecase_count + long_lowercase_count),
        -len(tokens),
        query.casefold(),
    )


def _lead_signal_penalty(item: dict[str, object]) -> int:
    if bool(item.get("materialized")):
        return 0
    locator = str(item.get("locator") or item.get("targetLocator") or "").strip()
    if not locator.startswith(("http://", "https://")):
        return 0
    provider = str(item.get("provider") or "")
    kind = str(item.get("kind") or "")
    if provider in FIRST_PARTY_LEAD_PROVIDERS and kind != "url":
        return 0
    return low_signal_url_penalty(locator)



#!/usr/bin/env python3
"""CLI session-root resolution and bootstrap policy."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from gotta.builtin import SessionAccessMode
import gotta.cli.argv as cli_argv
import gotta.cli.bind as cli_bind
import gotta.cli.select as cli_select


@dataclass(frozen=True, slots=True)
class ResolvedSessionTarget:
    root: Path | None
    requested_root: Path | None
    session_access: SessionAccessMode
    explicit_actor: str | None
    shared_root_command: bool
    created: bool
    bound_current_context: bool


def resolve_session_target(
    argv: list[str],
    *,
    context_id: str,
    context_source: str,
) -> ResolvedSessionTarget:
    plugin_name = argv[0] if argv else ""
    explicit_session = cli_argv._explicit_session_arg(argv)
    explicit_actor = cli_argv._explicit_actor_arg(argv)
    explicit_target = bool(explicit_session or explicit_actor)
    session_access = cli_argv._session_access_mode(argv)
    auto_bootstrap = cli_select._should_auto_bootstrap_session(
        argv=argv,
        context_source=context_source,
        explicit_session=explicit_session,
        session_access=session_access,
    )
    shared_root_command = cli_select._uses_shared_session_root(
        argv=argv,
        explicit_session=explicit_session,
        explicit_actor=explicit_actor,
        session_access=session_access,
    )
    init_command = plugin_name == "session" and len(argv) >= 2 and argv[1] == "init"

    root = cli_select.resolve_requested_root(
        argv,
        context_id=context_id,
        explicit_session=explicit_session,
        explicit_actor=explicit_actor,
        session_access=session_access,
        shared_root_command=shared_root_command,
        init_command=init_command,
    )
    root, created, bound_current_context = cli_bind.resolve_target_root(
        root,
        context_id=context_id,
        context_source=context_source,
        session_access=session_access,
        auto_bootstrap=auto_bootstrap,
    )
    if root is None and session_access != "ambient":
        raise SystemExit(
            "this command requires an existing session; run `gotta session bind` "
            "first or pass `--session <session-id>`"
        )

    requested_root = root
    root = cli_select.ambient_target_root(
        root,
        session_access=session_access,
        shared_root_command=shared_root_command,
        explicit_target=explicit_target,
    )
    root, scaffold_created, bound_current_context = cli_bind.finalize_target_root(
        root,
        context_id=context_id,
        context_source=context_source,
        session_access=session_access,
        auto_bootstrap=auto_bootstrap,
        init_command=init_command,
        bound_current_context=bound_current_context,
    )
    if session_access == "read":
        root = cli_select.finalize_read_target_root(
            root,
            explicit_session=explicit_session,
            explicit_actor=explicit_actor,
            shared_root_command=shared_root_command,
        )
    return ResolvedSessionTarget(
        root=root,
        requested_root=requested_root,
        session_access=session_access,
        explicit_actor=explicit_actor,
        shared_root_command=shared_root_command,
        created=created or scaffold_created,
        bound_current_context=bound_current_context,
    )



"""Entrypoint for `gotta session analyze`."""

from __future__ import annotations

import argparse
import json
from typing import Any

from ..parse import explicit_session_ref, require_started_session, session_dirs_for_read
from ..scan.payload import scan_payload
from .lineage import lineage_focus_payload, lineage_payload
from .overview import (
    analysis_overview_payload,
    combined_analysis_payload,
)
from .render import (
    render_analysis_focus_text,
    render_analysis_mermaid,
    render_analysis_overview_markdown,
    render_analysis_overview_text,
    render_combined_focus_markdown,
    render_lineage_focus_markdown_section,
    render_lineage_focus_text,
    render_lineage_overview_markdown,
    render_lineage_overview_text,
    render_semantic_focus_markdown_section,
    render_semantic_mermaid,
    render_semantic_overview_markdown,
    render_semantic_overview_text,
    render_single_focus_markdown,
    render_text_bundle,
)
from .semantic import semantic_focus_payload, semantic_payload


def cmd_analyze(args: argparse.Namespace) -> int:
    dirs = session_dirs_for_read(args)
    require_started_session(dirs)
    session_ref = explicit_session_ref(args)
    lineage_graph = lineage_payload(dirs, session_ref=session_ref)
    lineage_mermaid = render_analysis_mermaid(lineage_graph)
    semantic_graph = semantic_payload(lineage_graph)
    semantic_mermaid = render_semantic_mermaid(semantic_graph)
    focus_query = str(getattr(args, "focus", "") or "").strip()
    focus_limit = max(int(getattr(args, "limit", 8) or 0), 0)
    overview = analysis_overview_payload(
        lineage_graph,
        semantic_graph,
        limit=focus_limit,
    )
    lineage_focus: dict[str, Any] | None = None
    semantic_focus: dict[str, Any] | None = None
    if focus_query:
        focus_scan = scan_payload(
            dirs,
            query=focus_query,
            limit=max(focus_limit * 2, 12),
            include_all=True,
            session_ref=session_ref,
        )
        lineage_focus = lineage_focus_payload(
            lineage_graph,
            focus=focus_query,
            limit=focus_limit,
            scan_payload=focus_scan,
        )
        semantic_focus = semantic_focus_payload(
            lineage_graph,
            semantic_graph,
            focus=focus_query,
            limit=focus_limit,
            scan_payload=focus_scan,
        )

    if args.output == "text":
        if args.mode == "lineage":
            print(
                render_lineage_focus_text(lineage_focus)
                if lineage_focus is not None
                else render_lineage_overview_text(lineage_graph, limit=focus_limit)
            )
        elif args.mode == "semantic":
            print(
                render_analysis_focus_text(semantic_focus)
                if semantic_focus is not None
                else render_semantic_overview_text(overview)
            )
        elif focus_query:
            assert lineage_focus is not None
            assert semantic_focus is not None
            print(
                render_text_bundle(
                    [
                        ("Lineage", render_lineage_focus_text(lineage_focus)),
                        ("Semantic", render_analysis_focus_text(semantic_focus)),
                    ]
                )
            )
        else:
            print(render_analysis_overview_text(overview))
    elif args.output == "json":
        if args.mode == "lineage":
            print(
                json.dumps(
                    lineage_focus if lineage_focus is not None else lineage_graph,
                    indent=2,
                    sort_keys=True,
                )
            )
        elif args.mode == "semantic":
            print(
                json.dumps(
                    semantic_focus if semantic_focus is not None else semantic_graph,
                    indent=2,
                    sort_keys=True,
                )
            )
        else:
            print(
                json.dumps(
                    combined_analysis_payload(
                        focus=focus_query,
                        lineage=(
                            lineage_focus
                            if lineage_focus is not None
                            else lineage_graph
                        ),
                        semantic=(
                            semantic_focus
                            if semantic_focus is not None
                            else semantic_graph
                        ),
                    ),
                    indent=2,
                    sort_keys=True,
                )
            )
    elif args.output == "markdown":
        if args.mode == "lineage":
            print(
                render_single_focus_markdown(
                    session_dir=str(lineage_focus["sessionDir"]),
                    focus=str(lineage_focus.get("focus") or ""),
                    section_lines=render_lineage_focus_markdown_section(lineage_focus),
                )
                if lineage_focus is not None
                else render_lineage_overview_markdown(lineage_graph, limit=focus_limit),
                end="",
            )
        elif args.mode == "semantic":
            print(
                render_single_focus_markdown(
                    session_dir=str(semantic_focus["sessionDir"]),
                    focus=str(semantic_focus.get("focus") or ""),
                    section_lines=render_semantic_focus_markdown_section(
                        semantic_focus
                    ),
                )
                if semantic_focus is not None
                else render_semantic_overview_markdown(overview),
                end="",
            )
        elif focus_query:
            assert lineage_focus is not None
            assert semantic_focus is not None
            print(
                render_combined_focus_markdown(
                    lineage=lineage_focus,
                    semantic=semantic_focus,
                ),
                end="",
            )
        else:
            print(
                render_analysis_overview_markdown(
                    overview,
                    lineage=lineage_graph,
                    limit=focus_limit,
                ),
                end="",
            )
    else:
        if args.mode == "lineage":
            print(
                render_analysis_mermaid(lineage_focus)
                if lineage_focus is not None
                else lineage_mermaid
            )
        else:
            print(
                render_semantic_mermaid(semantic_focus)
                if semantic_focus is not None
                else semantic_mermaid
            )
    return 0



"""Canonicalize lead targets and infer providers."""

from __future__ import annotations

import re
import urllib.parse

from gotta.providers import atlassian as atl

SLACK_PERMALINK_RE = re.compile(
    r"https://[^/.]+\.slack\.com/archives/(?P<channel>[A-Z0-9]+)(?:/p(?P<pnum>[0-9]{16}))?"
)
JIRA_BROWSE_RE = re.compile(r"/browse/(?P<issue>[A-Z][A-Z0-9]+-\d+)(?:/|$)")
GDOC_URL_RE = re.compile(r"/document/d/(?P<doc_id>[A-Za-z0-9_-]+)(?:/|$)")
GSHEET_URL_RE = re.compile(r"/spreadsheets/d/(?P<sheet_id>[A-Za-z0-9_-]+)(?:/|$)")
GDRIVE_FILE_RE = re.compile(r"/file/d/(?P<file_id>[A-Za-z0-9_-]+)(?:/|$)")
LOW_SIGNAL_WEB_HOSTS = {
    "127.0.0.1",
    "localhost",
    "www.example.com",
    "example.com",
    "img.shields.io",
}
LOW_SIGNAL_WEB_EXTENSIONS = (".svg", ".png", ".jpg", ".jpeg", ".gif", ".webp")
LOW_SIGNAL_MEETING_HOSTS = {
    "meet.google.com",
}
LOW_SIGNAL_HOST_LABELS = {
    "admin",
    "argocd",
    "auth",
    "grafana",
    "jenkins",
    "login",
    "oauth",
    "oidc",
    "sso",
}
LOW_SIGNAL_PATH_MARKERS = (
    "/.well-known/",
    "/auth",
    "/explore",
    "/jwks",
    "/login",
    "/logout",
    "/oauth",
    "/oidc",
    "/openid",
    "/saml",
    "/signin",
    "/sso",
)


def provider_for_url(target: str) -> str:
    try:
        host = urllib.parse.urlparse(target).netloc.strip().lower()
    except ValueError:
        return "web"
    if "github.com" in host:
        return "github"
    if ".slack.com" in host or "enterprise.slack.com" in host:
        return "slack"
    if ".atlassian.net" in host:
        if "/wiki/" in target:
            return "confluence"
        return "jira"
    if "docs.google.com" in host:
        path = urllib.parse.urlparse(target).path
        if "/spreadsheets/" in path:
            return "gsheets"
        return "gdocs"
    if "drive.google.com" in host:
        return "gdrive"
    return "web"


def github_repo_reference(target: str) -> tuple[str, str]:
    try:
        parsed = urllib.parse.urlparse(target)
    except ValueError:
        return ("", "")
    if parsed.netloc.strip().lower() != "github.com":
        return ("", "")
    parts = [part for part in parsed.path.split("/") if part]
    if len(parts) < 2:
        return ("", "")
    repo_ref = f"{parts[0]}/{parts[1]}"
    if len(parts) == 2:
        return (repo_ref, "repo")
    if parts[2] == "tree" and len(parts) >= 4:
        return (repo_ref, "tree")
    if parts[2] == "blob" and len(parts) >= 4:
        return (repo_ref, "blob")
    return (repo_ref, "other")


def provider_for_locator(target: str) -> str:
    if target.startswith(("http://", "https://")):
        return provider_for_url(target)
    prefix = target.split(":", 1)[0].strip()
    if prefix in {
        "jira",
        "confluence",
        "gdocs",
        "gdrive",
        "gsheets",
        "slack",
        "artifact",
        "content",
    }:
        return prefix
    return "external"


def canonicalize_url(target: str) -> str | None:
    try:
        parsed = urllib.parse.urlparse(target)
    except ValueError:
        return target
    host = parsed.netloc.strip().lower()
    path = parsed.path
    query = parsed.query
    if host in {"www.google.com", "google.com"} and path == "/url":
        params = urllib.parse.parse_qs(query, keep_blank_values=True)
        for key in ("q", "url"):
            redirect = str((params.get(key) or [""])[0] or "").strip()
            if redirect:
                resolved = canonicalize_url(redirect)
                return resolved if resolved is not None else redirect
        filtered = [
            (key, value)
            for key, value in urllib.parse.parse_qsl(query, keep_blank_values=True)
            if key not in {"ust", "usg"}
        ]
        return urllib.parse.urlunsplit(
            (
                parsed.scheme,
                parsed.netloc,
                parsed.path,
                urllib.parse.urlencode(filtered, doseq=True),
                parsed.fragment,
            )
        )
    if _is_low_signal_web_url(target, host=host, path=path):
        return None
    if ".slack.com" in host:
        query_thread_ts = ""
        for value in urllib.parse.parse_qs(query).get("thread_ts") or []:
            candidate = str(value or "").strip()
            if re.fullmatch(r"\d{10}\.\d{6}", candidate):
                query_thread_ts = candidate.replace(".", "")
                break
        match = SLACK_PERMALINK_RE.match(target)
        if match:
            channel = match.group("channel")
            pnum = match.group("pnum")
            if query_thread_ts:
                return f"slack:thread:{channel}:{query_thread_ts}"
            if pnum:
                return f"slack:thread:{channel}:{pnum}"
            return f"slack:channel:{channel}"
    if ".atlassian.net" in host:
        if path.startswith("/wiki") and not re.match(r"^/wiki(?:/|$)", path):
            return None
        issue_match = JIRA_BROWSE_RE.search(path)
        if issue_match:
            return f"jira:{issue_match.group('issue')}"
        if "/browse/" in path:
            return None
        page_id = atl.extract_confluence_page_id(target)
        if page_id:
            return f"confluence:{page_id}"
    if "docs.google.com" in host:
        doc_match = GDOC_URL_RE.search(path)
        if doc_match:
            return f"gdocs:{doc_match.group('doc_id')}"
        sheet_match = GSHEET_URL_RE.search(path)
        if sheet_match:
            return f"gsheets:{sheet_match.group('sheet_id')}"
    if "drive.google.com" in host:
        drive_match = GDRIVE_FILE_RE.search(path)
        if drive_match:
            return f"gdrive:{drive_match.group('file_id')}"
    return target


def low_signal_url_penalty(locator: str) -> int:
    try:
        parsed = urllib.parse.urlparse(locator)
    except ValueError:
        return 0
    host = parsed.netloc.strip().lower()
    if not host:
        return 0
    host_only = host.split(":", 1)[0]
    path = parsed.path.strip().lower()
    labels = {part for part in re.split(r"[.-]", host_only) if part}
    if host_only in LOW_SIGNAL_MEETING_HOSTS or host_only.endswith(".zoom.us"):
        return 3
    if any(marker in path for marker in LOW_SIGNAL_PATH_MARKERS):
        return 3
    if labels & LOW_SIGNAL_HOST_LABELS and path in {"", "/"}:
        return 2
    if path in {"", "/"}:
        return 1
    return 0


def _is_low_signal_web_url(target: str, *, host: str, path: str) -> bool:
    host_only = host.split(":", 1)[0]
    lowered = target.strip().lower()
    if host_only in LOW_SIGNAL_WEB_HOSTS:
        return True
    if host_only.endswith(".example.com"):
        return True
    if "[server_addr]" in lowered or "x.x.x.x" in lowered:
        return True
    if path.lower().endswith(LOW_SIGNAL_WEB_EXTENSIONS):
        return True
    return False



"""Provider-owned locator routing helpers."""

from __future__ import annotations

from typing import Any

from gotta.builtin import PluginSpec, available_plugins, get_plugin
from gotta.content.model import CommonOptions
from gotta.content.path import sanitize_name
from gotta.resolve.model import ReadRequest, ReadTarget
import shlex
import urllib.parse


def split_locator_tail(tail: str) -> list[str]:
    try:
        parts = shlex.split(tail)
    except ValueError:
        parts = [tail]
    return [part for part in parts if part]


def strip_http_url_fragment(target: str) -> str:
    parsed = urllib.parse.urlparse(target)
    if parsed.scheme not in {"http", "https"} or not parsed.fragment:
        return target
    return urllib.parse.urlunparse(parsed._replace(fragment=""))


def query_route(
    subcommand: str,
    tail: str,
    *,
    valued_flags: tuple[str, ...] = (),
    boolean_flags: tuple[str, ...] = (),
) -> list[str] | None:
    parts = split_locator_tail(tail)
    if not parts:
        return None
    argv = [subcommand]
    query_parts: list[str] = []
    index = 0
    parse_flags = True
    while index < len(parts):
        token = parts[index]
        if parse_flags and token == "--":
            parse_flags = False
            index += 1
            continue
        if parse_flags and token in boolean_flags:
            argv.append(token)
            index += 1
            continue
        if parse_flags and token in valued_flags:
            if index + 1 >= len(parts):
                return None
            argv.extend([token, parts[index + 1]])
            index += 2
            continue
        if parse_flags:
            matched_flag = next(
                (flag for flag in valued_flags if token.startswith(f"{flag}=")),
                "",
            )
            if matched_flag:
                argv.extend([matched_flag, token.split("=", 1)[1]])
                index += 1
                continue
        query_parts.append(token)
        index += 1
    query = " ".join(part for part in query_parts if part).strip()
    if not query:
        return None
    return [*argv, query]


def _routed_plugins() -> list[PluginSpec]:
    return sorted(
        [
            spec
            for spec in (get_plugin(name) for name in available_plugins())
            if spec
            and spec.name not in {"read", "session"}
            and spec.route_target is not None
        ],
        key=lambda item: (item.route_priority, item.name),
    )


def discover_plugin_route(target: str) -> tuple[str, list[str]] | None:
    for plugin in _routed_plugins():
        try:
            argv = plugin.route_target(target) if plugin.route_target else None
        except ValueError:
            argv = None
        if argv is not None:
            return plugin.name, argv
    return None


def _resolved_routed_target(
    request: ReadRequest,
    target: str,
    plugin: str,
    plugin_argv: list[str],
    options: CommonOptions | Any,
    *,
    save_as: str,
) -> ReadTarget:
    spec = get_plugin(plugin)
    canonical = (
        spec.canonical_locator(plugin_argv)
        if spec and spec.canonical_locator
        else target
    )
    preferred = save_as or (
        spec.preferred_name(plugin_argv, options)
        if spec and spec.preferred_name
        else f"{sanitize_name(target) or plugin}.txt"
    )
    return ReadTarget(
        request=request,
        kind="routed",
        path=None,
        routed_plugin=plugin,
        routed_argv=plugin_argv,
        canonical_locator=canonical,
        preferred_name=preferred,
        should_materialize=True,
    )


def resolve_routed_target(
    request: ReadRequest,
    target: str,
    options: CommonOptions | Any,
    *,
    save_as: str,
) -> ReadTarget | None:
    if request.routed_plugin and request.routed_argv:
        return _resolved_routed_target(
            request,
            target,
            request.routed_plugin,
            list(request.routed_argv),
            options,
            save_as=save_as,
        )
    routed = discover_plugin_route(target)
    if routed is None:
        return None
    plugin, plugin_argv = routed
    return _resolved_routed_target(
        request,
        target,
        plugin,
        plugin_argv,
        options,
        save_as=save_as,
    )


def partition_routed_target_tokens(
    tokens: list[str],
) -> tuple[str, str, tuple[str, ...]] | None:
    best: tuple[int, int, str, str, tuple[str, ...]] | None = None
    token_count = len(tokens)
    for start in range(token_count):
        for end in range(token_count, start, -1):
            candidate = " ".join(
                part.strip() for part in tokens[start:end] if part.strip()
            ).strip()
            if not candidate:
                continue
            routed = discover_plugin_route(candidate)
            if routed is None:
                continue
            plugin, plugin_argv = routed
            provider_argv = tuple(tokens[:start] + plugin_argv + tokens[end:])
            score = (end - start, -len(tokens[:start] + tokens[end:]))
            if best is None or score > (best[0], best[1]):
                best = (score[0], score[1], plugin, candidate, provider_argv)
    if best is None:
        return None
    return best[2], best[3], best[4]



"""Doctor surface for `gotta session`."""

from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import TypedDict

from gotta.content.context import current_context_binding
from gotta.content.env import SESSION_CREATED_ENV, load_state_env_at_root
from gotta.content.model import ResolvedDirs
from gotta.content.scope import session_is_initialized, shared_session_root
from gotta import topology

from .parse import require_started_session, session_dirs_for_read, session_scope_started
from .show import show_payload


class DoctorRuntimePayload(TypedDict):
    present: bool
    contextId: str
    contextSource: str
    bindingId: str


class DoctorSessionPayload(TypedDict):
    sessionId: str
    actor: str
    sessionRoot: str
    contentRoot: str
    initialized: bool
    repo: str
    createdAt: str


class DoctorCheck(TypedDict):
    status: str
    detail: str


class DoctorChecks(TypedDict):
    runtimeContextPresent: DoctorCheck
    durableBindingsPresent: DoctorCheck
    runtimeBindingMatchesTarget: DoctorCheck
    sessionTopologyConsistent: DoctorCheck


class DoctorPayload(TypedDict):
    runtime: DoctorRuntimePayload
    session: DoctorSessionPayload
    bindings: list[dict[str, object]]
    checks: DoctorChecks


def _binding_detail(record: dict[str, object]) -> str:
    context_id = str(record.get("contextId") or "").strip() or "unknown"
    context_source = str(record.get("contextSource") or "").strip() or "unknown"
    session_root = str(record.get("sessionRoot") or "").strip() or "unknown"
    return (
        f"{record.get('bindingId') or 'unknown'} -> {session_root} "
        f"({context_source}, {context_id})"
    )


def doctor_payload(dirs: ResolvedDirs) -> DoctorPayload:
    session_env = show_payload(dirs)
    state = load_state_env_at_root(dirs.session_dir)
    runtime = current_context_binding()
    runtime_root = topology.resolve_binding(runtime.binding_id)
    bindings = topology.binding_records_for_session(dirs.session_dir)
    target_session_id = str(session_env.get("GOTTA_SESSION_ID") or "")
    target_shared_root = shared_session_root(target_session_id).resolve()
    session_payload: DoctorSessionPayload = {
        "sessionId": target_session_id,
        "actor": str(session_env.get("GOTTA_SESSION_ACTOR") or ""),
        "sessionRoot": str(dirs.session_dir),
        "contentRoot": str(dirs.content_dir),
        "initialized": bool(session_is_initialized(dirs.session_dir)),
        "repo": str(state.get("GOTTA_SESSION_REPO") or ""),
        "createdAt": str(state.get(SESSION_CREATED_ENV) or ""),
    }
    runtime_payload: DoctorRuntimePayload = {
        "present": bool(runtime.context_id),
        "contextId": runtime.context_id,
        "contextSource": runtime.context_source,
        "bindingId": runtime.binding_id,
    }
    session_in_shared_topology = (
        topology.parse_grouped_session_root(dirs.session_dir.resolve()) is not None
        or topology.parse_shared_session_root(dirs.session_dir.resolve()) is not None
    )
    matching_runtime_binding = (
        runtime_root is not None
        and topology.shared_session_id(runtime_root) == target_session_id
    )
    bindings_match_target = all(
        topology.normalize_session_id(str(record.get("sessionId") or ""))
        == target_session_id
        and topology.shared_session_id(
            Path(str(record.get("sessionRoot") or target_shared_root))
        )
        == target_session_id
        for record in bindings
    )
    topology_consistent = (
        dirs.session_dir.exists()
        and dirs.content_dir.exists()
        and session_scope_started(dirs)
        and topology.shared_session_id(dirs.session_dir) == target_session_id
        and (
            not session_in_shared_topology
            or dirs.content_dir.resolve() == (target_shared_root / "content").resolve()
        )
        and bool(bindings)
        and bindings_match_target
    )
    checks: DoctorChecks = {
        "runtimeContextPresent": {
            "status": "ok" if runtime_payload["present"] else "missing",
            "detail": (
                f"{runtime.context_source}: {runtime.context_id}"
                if runtime_payload["present"]
                else "no active runtime context"
            ),
        },
        "durableBindingsPresent": {
            "status": "ok" if bindings else "missing",
            "detail": (
                ", ".join(_binding_detail(record) for record in bindings)
                if bindings
                else "no durable bindings target this session"
            ),
        },
        "runtimeBindingMatchesTarget": {
            "status": (
                "ok"
                if matching_runtime_binding
                else "mismatch"
                if runtime_payload["present"]
                else "unknown"
            ),
            "detail": (
                f"{runtime.binding_id} targets session "
                f"{topology.shared_session_id(runtime_root)} at {runtime_root}"
                if runtime_root is not None
                else "the active runtime binding has no durable target"
            )
            if runtime_payload["present"]
            else "no active runtime context",
        },
        "sessionTopologyConsistent": {
            "status": "ok" if topology_consistent else "broken",
            "detail": (
                "session root, content root, and binding records agree"
                if topology_consistent
                else "session root, content root, and durable binding records do not fully agree"
            ),
        },
    }
    return {
        "runtime": runtime_payload,
        "session": session_payload,
        "bindings": bindings,
        "checks": checks,
    }


def _print_doctor_summary(payload: DoctorPayload) -> int:
    runtime = payload["runtime"]
    session_payload = payload["session"]
    print(f"session: {session_payload['sessionRoot']}")
    print(f"session_id: {session_payload['sessionId']}")
    print(f"actor: {session_payload['actor']}")
    print(f"content: {session_payload['contentRoot']}")
    print(
        "runtime: "
        f"{runtime['contextSource'] or 'unknown'} "
        f"{runtime['contextId'] or 'unknown'} "
        f"({runtime['bindingId'] or 'unknown'})"
    )
    print(f"bindings: {len(payload['bindings'])}")
    for name in (
        "runtimeContextPresent",
        "durableBindingsPresent",
        "runtimeBindingMatchesTarget",
        "sessionTopologyConsistent",
    ):
        check = payload["checks"][name]
        print(f"- {name}: {check['status']} - {check['detail']}")
    return 0


def cmd_doctor(args: argparse.Namespace) -> int:
    dirs = session_dirs_for_read(args)
    require_started_session(dirs)
    payload = doctor_payload(dirs)
    if getattr(args, "print_format", "json") == "json":
        print(json.dumps(payload, indent=2, sort_keys=True))
        return 0
    return _print_doctor_summary(payload)



"""Mutation paths for `gotta logs`."""

from __future__ import annotations

import argparse
from pathlib import Path

from gotta.actor import require_writer, session_actor, writer_name
from gotta.logs import append_log_record, logs_state_path
from gotta.session import charter as session_charter
from gotta.session import scope as session_scope
from gotta.session.activity.record import _record_session_activity


def _write_target(args: argparse.Namespace) -> tuple[Path, str]:
    work_dir = session_scope._session_dir(
        explicit_session=getattr(args, "session", None),
        explicit_actor=getattr(args, "actor", None),
    )
    writer = writer_name()
    target_actor = (
        session_actor(work_dir) if work_dir.resolve().parent.name == "actors" else ""
    )
    if target_actor:
        require_writer(
            work_dir,
            target_actor,
            writer=writer,
            action="write into this actor branch",
        )
    return work_dir, writer or ""


def extend_logs(args: argparse.Namespace) -> int:
    work_dir, record_actor = _write_target(args)
    entries = session_charter._read_text_items_source(
        session_root=work_dir,
        inline_items=list(args.value or []),
        from_file=args.from_file,
        use_stdin=args.use_stdin,
        input_name="log entry text",
    )
    for entry in entries:
        append_log_record(
            work_dir,
            message=session_charter._normalize_entry_text(
                entry, input_name="log entry text"
            ),
            actor=record_actor,
        )
    _record_session_activity(
        work_dir,
        plugin="logs",
        surface="logs",
        action="extend",
        actor=record_actor,
        target=logs_state_path(work_dir),
        detail=f"extended logs with {len(entries)} item(s)",
    )
    print(f"extended logs entries: {len(entries)} item(s)")
    return 0


def append_logs(args: argparse.Namespace) -> int:
    work_dir, record_actor = _write_target(args)
    payload = session_charter._read_text_source(
        session_root=work_dir,
        inline=(args.value[0] if args.value else None),
        from_file=args.from_file,
        use_stdin=args.use_stdin,
        input_name="log entry text",
    )
    append_log_record(
        work_dir,
        message=session_charter._normalize_entry_text(
            payload, input_name="log entry text"
        ),
        actor=record_actor,
    )
    _record_session_activity(
        work_dir,
        plugin="logs",
        surface="logs",
        action="append",
        actor=record_actor,
        target=logs_state_path(work_dir),
        detail="appended 1 logs entry",
    )
    print("appended logs entry")
    return 0



"""Activity, evidence, and note-state payload helpers."""

from __future__ import annotations

from pathlib import Path

from gotta.notes.voice import actor_notes_status, actor_voice
from gotta.session.activity.summary import (
    _actor_evidence_note,
    _actor_evidence_summary,
    _actor_note_check_summary,
    _actor_note_summary,
    _actor_recent_activity,
)
from gotta.session.registry import (
    _actor_session_dir,
    _normalize_actor_name,
)
from gotta.session.status.payload.value import (
    ACTOR_TERMINAL_STATUS,
    int_value,
    lifecycle_entries,
)
from gotta.session.status.progress import _actor_progress_summary


def _normalized_recent_activity(
    recent_activity: dict[str, object],
) -> dict[str, object]:
    payload = dict(recent_activity)
    entries = lifecycle_entries(payload.get("recent_lifecycle"))
    if entries and str(entries[0].get("event") or "") == "runtime_exit":
        lifecycle_detail = str(entries[0].get("detail") or "")
        request_labels = {
            "runtime_stop_requested": "runtime stop signal",
            "failed_requested": "failure request",
            "signoff_requested": "sign-off request",
            "complete_requested": "completion request",
        }
        honored = next(
            (
                request_labels.get(str(item.get("event") or ""))
                for item in entries[1:]
                if str(item.get("event") or "") in request_labels
            ),
            "",
        )
        if honored and "code 0" in lifecycle_detail:
            entries[0]["summary"] = (
                f"runtime exit: actor process exited cleanly after honoring {honored}"
            )
    if entries:
        payload["recent_lifecycle"] = entries
        payload["last_lifecycle_at"] = str(entries[0].get("timestamp") or "")
        payload["last_lifecycle_summary"] = str(entries[0].get("summary") or "")
    return payload


def activity_payload(work_dir: Path, actor_name: str) -> dict[str, object]:
    normalized_actor = _normalize_actor_name(actor_name)
    evidence = _actor_evidence_summary(work_dir, actor_name)
    notes_status = actor_notes_status(work_dir, normalized_actor)
    return {
        "actor_dir": _actor_session_dir(work_dir, actor_name),
        "voice": actor_voice(work_dir, normalized_actor),
        "notes_status": notes_status,
        "notes_ready": notes_status == "present",
        "evidence": evidence,
        "evidence_note": _actor_evidence_note(evidence),
        "evidence_live": int_value(evidence.get("artifact_count")) > 0,
        "recent_activity": _normalized_recent_activity(
            _actor_recent_activity(work_dir, actor_name)
        ),
        "note_summary": _actor_note_summary(work_dir, actor_name),
        "note_check_summary": _actor_note_check_summary(work_dir, actor_name),
        "progress": _actor_progress_summary(work_dir, actor_name),
    }


def note_check_next_step(
    *,
    note_checks_since_update: int,
    last_note_at: str,
    derived_status: str,
) -> str:
    if note_checks_since_update <= 0 or derived_status in ACTOR_TERMINAL_STATUS:
        return ""
    check_noun = "time" if note_checks_since_update == 1 else "times"
    if last_note_at:
        return (
            f"Supervisor has checked this actor's notes {note_checks_since_update} {check_noun} "
            "since the last note. Land one short note now if real progress exists."
        )
    return (
        f"Supervisor has checked this actor's notes {note_checks_since_update} {check_noun} "
        "and no first short note has landed yet. Land one short note now if real progress exists."
    )



"""Session activity and actor event recording helpers."""

from __future__ import annotations

from pathlib import Path

from gotta.compat import UTC, datetime
from gotta.content.activity import append_activity_event
from gotta.content.context import current_actor
from gotta.content.scope import session_identity
from gotta.friction import oops_log_path
from gotta.logs import append_log_record, logs_state_path
from gotta.todo import todo_state_path

from gotta.session.charter import (
    _native_surface_follow_command,
    _native_surface_locator,
    _native_surface_preferred_name,
    _session_relative_locator,
    _surface_actor_scope,
)
from gotta.session.registry import (
    WANT_FILE,
    _actor_events_path,
    _normalize_actor_name,
)

from .file import _append_jsonl


def _record_session_activity(
    work_dir: Path,
    *,
    plugin: str,
    surface: str,
    action: str,
    actor: str = "",
    target_actor: str = "",
    target: Path | None = None,
    locator: str = "",
    preferred_name: str = "",
    follow_command: str = "",
    detail: str = "",
) -> None:
    if target is not None:
        resolved = target.resolve()
        actor_scope = _surface_actor_scope(work_dir)
        if resolved == (work_dir / WANT_FILE).resolve():
            resolved_locator = _native_surface_locator("want", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "want", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "want", actor_name=actor_scope
            )
        elif resolved == (work_dir / "GOAL.md").resolve():
            resolved_locator = _native_surface_locator("goal", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "goal", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "goal", actor_name=actor_scope
            )
        elif resolved == todo_state_path(work_dir).resolve():
            resolved_locator = _native_surface_locator("todo", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "todo", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "todo", actor_name=actor_scope
            )
        elif resolved == logs_state_path(work_dir).resolve():
            resolved_locator = _native_surface_locator("logs", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "logs", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "logs", actor_name=actor_scope
            )
        elif resolved == oops_log_path(work_dir).resolve():
            resolved_locator = _native_surface_locator("oops", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "oops", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "oops", actor_name=actor_scope
            )
        else:
            resolved_locator = _session_relative_locator(work_dir, resolved)
            resolved_name = resolved.name
            resolved_follow = f"gotta read {resolved_locator!r}"
    else:
        resolved_locator = locator.strip() or f"{plugin}:{surface}"
        resolved_name = preferred_name.strip() or resolved_locator
        resolved_follow = follow_command.strip()
    activity_actor = actor.strip() or current_actor(
        default_actor=session_identity(work_dir)
    )
    payload = {
        "plugin": plugin,
        "surface": surface,
        "action": action,
        "actor": activity_actor,
        "locator": resolved_locator,
        "preferred_name": preferred_name.strip() or resolved_name,
        "follow_command": follow_command.strip() or resolved_follow,
        "detail": detail,
        "time_field": "session_recorded_at",
    }
    normalized_target = (
        _normalize_actor_name(target_actor) if target_actor.strip() else ""
    )
    if normalized_target and normalized_target != activity_actor:
        payload["target_actor"] = normalized_target
    append_activity_event(work_dir, payload)


def _append_actor_event(
    work_dir: Path,
    actor_name: str,
    *,
    event: str,
    detail: str = "",
    extra: dict[str, object] | None = None,
    author: str = "",
) -> None:
    normalized_actor = _normalize_actor_name(actor_name)
    event_author = author.strip() or current_actor(default_actor=normalized_actor)
    payload: dict[str, object] = {
        "timestamp": datetime.now(tz=UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
        "actor": normalized_actor,
        "author": event_author,
        "event": event,
        "detail": detail,
    }
    if extra:
        payload.update(extra)
    _append_jsonl(_actor_events_path(work_dir, actor_name), payload)
    if event != "heartbeat":
        _record_session_activity(
            work_dir,
            plugin="actor",
            surface="actor.lifecycle",
            action=event,
            actor=event_author,
            target_actor=normalized_actor,
            locator=f"actor:{payload['actor']}",
            preferred_name=str(payload["actor"]),
            follow_command=f"gotta actor status {payload['actor']}",
            detail=detail or event,
        )


def _actor_log_line(
    session_root: Path, actor_name: str, message: str, *, author: str = ""
) -> None:
    normalized_actor = _normalize_actor_name(actor_name)
    log_author = author.strip() or current_actor(default_actor=normalized_actor)
    if log_author == normalized_actor:
        rendered = f"[{normalized_actor}] {message}"
    else:
        rendered = f"[{log_author} -> {normalized_actor}] {message}"
    append_log_record(session_root, message=rendered, actor=log_author)


def _record_actor_surface_activity(
    session_root: Path,
    *,
    actor_name: str,
    surface: str,
    action: str,
    detail: str,
    actor: str = "",
) -> None:
    normalized_actor = _normalize_actor_name(actor_name)
    _record_session_activity(
        session_root,
        plugin="actor",
        surface=surface,
        action=action,
        actor=actor.strip() or current_actor(default_actor=normalized_actor),
        target_actor=normalized_actor,
        locator=_native_surface_locator(surface, actor_name=normalized_actor),
        preferred_name=_native_surface_preferred_name(
            surface, actor_name=normalized_actor
        ),
        follow_command=_native_surface_follow_command(
            surface, actor_name=normalized_actor
        ),
        detail=detail,
    )



#!/usr/bin/env python3
"""CLI session/root selection policy."""

from __future__ import annotations

import os
from pathlib import Path

from gotta.builtin import SessionAccessMode
import gotta.cli.bind as cli_bind
import gotta.cli.env as cli_env
from gotta.content.context import current_context_binding, default_session_id
from gotta.content.env import SESSION_ACTOR_ENV, SESSION_ENV
from gotta.content.model import CommonOptions
from gotta.content.scope import resolve_session_reference, session_is_initialized
from gotta import topology
from gotta.session import registry as session_registry
from gotta.session import scope as session_scope

_AUTO_BOOTSTRAP_CONTEXT_SOURCES = {"codex_thread", "terminal_session"}


def _resolve_shared_explicit_session(raw: str) -> Path | None:
    normalized = raw.strip()
    if not normalized:
        return None
    candidate = Path(normalized).expanduser()
    if candidate.is_absolute():
        resolved = candidate.resolve()
        if topology.parse_shared_session_root(resolved) is not None:
            return resolved
        if topology.parse_grouped_session_root(resolved) is not None:
            return resolved
        if (resolved / "state" / "env").exists():
            return resolved
        return None
    if "/" in normalized:
        root = resolve_session_reference(normalized, allow_missing=False)
        if root is None:
            return None
        return root
    shared_root = topology.shared_session_root_for(normalized)
    if shared_root.exists() or shared_root.is_symlink():
        return shared_root.resolve()
    return None


def _prefer_bound_session_root() -> Path | None:
    explicit = os.environ.get(SESSION_ENV, "").strip()
    if explicit:
        root = resolve_session_reference(explicit, allow_missing=False)
        if root is not None and session_is_initialized(root):
            return root
    root = topology.resolve_binding(current_context_binding().binding_id)
    if root is not None and session_is_initialized(root):
        return root
    return None


def _active_identity(context_id: str) -> str:
    current = _prefer_bound_session_root()
    if current is not None:
        current_identity = topology.session_identity(current)
        if current_identity and not topology.is_placeholder_identity(current_identity):
            return current_identity
    explicit = os.environ.get(SESSION_ACTOR_ENV, "").strip()
    if explicit and not topology.is_placeholder_identity(explicit):
        return topology.normalize_identity(explicit)
    return default_session_id(context_id)


def _should_auto_bootstrap_session(
    *,
    argv: list[str],
    context_source: str,
    explicit_session: str | None,
    session_access: SessionAccessMode,
) -> bool:
    if explicit_session:
        return False
    if session_access == "none":
        return False
    if context_source not in _AUTO_BOOTSTRAP_CONTEXT_SOURCES:
        return False
    if argv[:1] != ["read"]:
        return True
    try:
        from gotta.resolve.read import resolve_read_target

        resolved = resolve_read_target(argv[1:])
    except SystemExit:
        return False
    return resolved.should_materialize


def _existing_actor_root_for_session(
    root: Path,
    *,
    preferred_identities: list[str],
) -> Path | None:
    actors_dir = session_registry._group_session_root(root) / "actors"
    if not actors_dir.is_dir():
        return None
    initialized: dict[str, Path] = {}
    for actor_dir in sorted(actors_dir.iterdir()):
        if not actor_dir.is_dir():
            continue
        resolved = actor_dir.resolve()
        if not session_is_initialized(resolved):
            continue
        identity = topology.session_identity(resolved)
        if not identity or topology.is_placeholder_identity(identity):
            continue
        initialized[identity] = resolved
    for identity in preferred_identities:
        normalized = topology.normalize_identity(identity)
        match = initialized.get(normalized)
        if match is not None:
            return match
    return next(iter(initialized.values()), None)


def _preferred_read_only_session_identities(
    root: Path,
    *,
    explicit_actor: str | None,
) -> list[str]:
    preferred: list[str] = []
    if explicit_actor:
        preferred.append(explicit_actor)
    current = _prefer_bound_session_root()
    if current is None:
        return preferred
    if topology.shared_session_id(current) != topology.shared_session_id(root):
        return preferred
    current_identity = topology.session_identity(current)
    if current_identity and not topology.is_placeholder_identity(current_identity):
        preferred.append(current_identity)
    return preferred


def _is_shared_session_root(root: Path | None) -> bool:
    if root is None:
        return False
    resolved = root.expanduser().resolve()
    return topology.parse_shared_session_root(resolved) is not None


def _looks_like_stored_read_target(argv: list[str], explicit_session: str) -> bool:
    if argv[:1] != ["read"]:
        return False
    shared_root = _resolve_shared_explicit_session(explicit_session)
    if shared_root is None:
        return False
    try:
        from gotta.resolve.read import resolve_read_target

        resolved = resolve_read_target(
            argv[1:],
            CommonOptions(
                session_dir=str(shared_root),
                content_dir=str(cli_env._explicit_session_content_root(shared_root)),
            ),
        )
    except SystemExit:
        return False
    target = str(resolved.request.target or "").strip()
    if resolved.kind in {"artifact_locator", "artifact_name"}:
        return True
    if target.startswith("content:"):
        return True
    return len(target) == 64 and all(ch in "0123456789abcdef" for ch in target)


def _uses_shared_session_root(
    *,
    argv: list[str],
    explicit_session: str | None,
    explicit_actor: str | None,
    session_access: SessionAccessMode,
) -> bool:
    if not explicit_session or explicit_actor:
        return False
    if argv[:1] == ["session"]:
        return True
    if argv[:1] == ["notes"]:
        return True
    if session_access == "ambient":
        return _looks_like_stored_read_target(argv, explicit_session)
    return False


def _prefers_primary_actor_root(argv: list[str]) -> bool:
    if not argv:
        return False
    plugin = argv[0]
    if plugin in {"want", "goal"}:
        return True
    if plugin == "session" and len(argv) >= 2 and argv[1] == "show":
        return True
    if plugin == "actor" and len(argv) >= 2 and argv[1] == "status":
        return True
    return False


def _resolve_primary_actor_root(root: Path) -> Path | None:
    session_root = root.expanduser().resolve()
    if not _is_shared_session_root(session_root):
        return session_root
    primary = session_scope._primary_actor_name(session_root)
    if not primary:
        return None
    return session_registry._actor_session_dir(session_root, primary)


def resolve_requested_root(
    argv: list[str],
    *,
    context_id: str,
    explicit_session: str | None,
    explicit_actor: str | None,
    session_access: SessionAccessMode,
    shared_root_command: bool,
    init_command: bool,
) -> Path | None:
    if explicit_session:
        root = resolve_explicit_session_root(
            argv,
            context_id=context_id,
            explicit_session=explicit_session,
            explicit_actor=explicit_actor,
            session_access=session_access,
            shared_root_command=shared_root_command,
            init_command=init_command,
        )
        if root is None:
            raise SystemExit(
                "session references must be an absolute path, a shared session id, "
                "or an explicit <session>/<actor> session reference"
            )
        return root
    if explicit_actor:
        current = _prefer_bound_session_root()
        if current is None:
            current = cli_bind._resolve_existing_session_root(context_id)
        if current is None:
            raise SystemExit(
                "explicit actor targeting requires an existing session; run "
                "`gotta session bind` first or pass `--session <session-id>`"
            )
        resolved_actor = session_registry._resolve_bound_actor_name(
            current,
            explicit_actor,
        )
        return session_registry._actor_session_dir(current, resolved_actor)
    return _prefer_bound_session_root()


def resolve_explicit_session_root(
    argv: list[str],
    *,
    context_id: str,
    explicit_session: str,
    explicit_actor: str | None,
    session_access: SessionAccessMode,
    shared_root_command: bool,
    init_command: bool,
) -> Path | None:
    if init_command:
        target_identity = topology.normalize_identity(_active_identity(context_id))
        return resolve_session_reference(
            explicit_session,
            identity=target_identity,
            allow_missing=True,
        )
    if shared_root_command:
        return _resolve_shared_explicit_session(explicit_session)
    if (
        session_access == "read"
        and not explicit_actor
        and _prefers_primary_actor_root(argv)
    ):
        shared_root = _resolve_shared_explicit_session(explicit_session)
        if shared_root is None or not _is_shared_session_root(shared_root):
            return shared_root
        primary_root = _resolve_primary_actor_root(shared_root)
        if primary_root is None:
            raise SystemExit(
                "this shared session does not resolve to one canonical actor root; "
                "pass `--actor <actor>` explicitly"
            )
        return primary_root
    target_identity = topology.normalize_identity(
        explicit_actor or _active_identity(context_id)
    )
    explicit_root = resolve_session_reference(
        explicit_session,
        identity=target_identity,
        allow_missing=False,
    )
    if explicit_root is not None and explicit_actor:
        resolved_actor = session_registry._resolve_bound_actor_name(
            explicit_root,
            explicit_actor,
        )
        return session_registry._actor_session_dir(explicit_root, resolved_actor)
    if explicit_root is not None:
        return explicit_root
    if explicit_actor:
        raise SystemExit(
            "explicit actor targeting requires an existing shared session and a bound actor"
        )
    return resolve_session_reference(
        explicit_session,
        identity=target_identity,
        allow_missing=True,
    )


def ambient_target_root(
    root: Path | None,
    *,
    session_access: SessionAccessMode,
    shared_root_command: bool,
    explicit_target: bool,
) -> Path | None:
    if (
        session_access == "ambient"
        and root is not None
        and not session_is_initialized(root)
        and not shared_root_command
    ):
        if explicit_target:
            raise SystemExit(
                "ambient retrieval requires an existing initialized actor root in the "
                "target session; bind an actor there first or pass `--actor <actor>`"
            )
        return None
    return root


def finalize_read_target_root(
    root: Path | None,
    *,
    explicit_session: str | None,
    explicit_actor: str | None,
    shared_root_command: bool,
) -> Path:
    if (
        root is not None
        and not session_is_initialized(root)
        and not shared_root_command
    ):
        existing = _existing_actor_root_for_session(
            root,
            preferred_identities=_preferred_read_only_session_identities(
                root,
                explicit_actor=explicit_actor,
            ),
        )
        if existing is not None:
            root = existing
    if root is None:
        raise SystemExit(
            "explicit session inspection requires an initialized actor root in the "
            "target shared session; bind an actor there first or pass --actor"
        )
    if session_is_initialized(root) or shared_root_command:
        return root
    if explicit_session and not _is_shared_session_root(root):
        raise SystemExit(
            "explicit session inspection requires an existing initialized session "
            "at that exact root"
        )
    raise SystemExit(
        "explicit session inspection requires an initialized actor root in the "
        "target shared session; bind an actor there first or pass --actor"
    )



"""Receipt emission and follow-command helpers for dispatch."""

from __future__ import annotations

import json
import os
import shlex
import sys
from typing import Any

from gotta.content.model import Materialization
from gotta.content.path import artifact_locator, content_locator
from gotta.dispatch.budget import EmittedOutput, _output_budget_descriptor


SUPPRESS_RECEIPTS_ENV = "GOTTA_SUPPRESS_RECEIPTS"
_HELP_TOKENS = {"-h", "--help", "--help-all"}


def _result_follow_command(result: Materialization | None) -> str:
    if result is None:
        return ""
    locator = content_locator(result.digest)
    return shlex.join(["gotta", "read", locator])


def _rerun_full_output_command(plugin: str, argv: list[str]) -> str:
    return shlex.join(["gotta", plugin, *argv, "--full-output"])


def _receipt_payload(
    *,
    emitted: EmittedOutput,
    result: Materialization | None,
    extra: dict[str, Any] | None = None,
) -> dict[str, Any]:
    payload: dict[str, Any] = {}
    if emitted.output_truncated:
        payload.update(
            {
                "outputFormat": emitted.format,
                "outputBudgetApplied": emitted.output_budget_applied,
                "outputTruncated": True,
                "truncateReason": emitted.truncate_reason or None,
                "budget": _output_budget_descriptor(),
                "originalBytes": emitted.original_bytes,
                "emittedBytes": emitted.emitted_bytes,
            }
        )
        if emitted.original_lines is not None:
            payload["originalLines"] = emitted.original_lines
        if emitted.emitted_lines is not None:
            payload["emittedLines"] = emitted.emitted_lines
    if result is not None:
        payload["artifactKind"] = str(result.artifact_kind or "").strip() or "content"
        payload["artifactLocator"] = artifact_locator(
            result.name_link.name, result.digest
        )
        payload["contentLocator"] = content_locator(result.digest)
        payload["followCommand"] = _result_follow_command(result)
    if extra:
        payload.update(extra)
    return payload


def _emit_receipt(payload: dict[str, Any], *, quiet: bool) -> None:
    if not payload or quiet or os.environ.get(SUPPRESS_RECEIPTS_ENV) == "1":
        return
    print(json.dumps(payload, sort_keys=True, separators=(",", ":")), file=sys.stderr)


def _should_emit_receipt(plugin: str, argv: list[str]) -> bool:
    if os.environ.get(SUPPRESS_RECEIPTS_ENV) == "1":
        return False
    if any(token in _HELP_TOKENS for token in argv):
        return False
    if plugin in {"ask"}:
        return False
    return True


def _receipt_extra(
    plugin: str,
    argv: list[str],
    *,
    dirs: object | None,
) -> dict[str, Any]:
    return {}



"""Show surface for `gotta session`."""

from __future__ import annotations

import argparse
import json

from gotta.content.model import ResolvedDirs
from gotta.content.env import CONTENT_ENV, SESSION_ENV
from gotta.content.path import sh_quote
from gotta.content.scope import session_identity, session_shared_id
from gotta import topology
from gotta.session import registry as session_registry
from gotta.session import scope as session_scope

from .parse import require_started_session, session_dirs_for_read

SessionEnvPayload = dict[str, str]


def print_session_env(
    args: argparse.Namespace,
    payload: SessionEnvPayload,
) -> int:
    print_format = getattr(args, "print_format", "path")
    if print_format == "json":
        print(json.dumps(payload, indent=2, sort_keys=True))
        return 0
    if print_format == "path":
        print(payload["GOTTA_SESSION_DIR"])
        return 0
    if print_format == "env":
        for key, value in payload.items():
            print(f"{key}={value}")
        return 0
    print(
        "\n".join(f"export {key}={sh_quote(value)}" for key, value in payload.items())
    )
    return 0


def show_payload(dirs: ResolvedDirs) -> SessionEnvPayload:
    resolved = dirs.session_dir.resolve()
    if resolved.parent.name == "actors":
        actor_root = resolved
        session_root = session_registry._group_session_root(actor_root)
        primary_actor = session_identity(actor_root)
        state_dir = str(actor_root / "state")
    elif topology.parse_shared_session_root(resolved) is None:
        actor_root = resolved
        session_root = session_registry._group_session_root(resolved)
        primary_actor = ""
        state_dir = str(actor_root / "state")
    else:
        session_root = session_registry._group_session_root(resolved)
        primary_actor = session_scope._primary_actor_name(session_root) or ""
        actor_root = (
            session_registry._actor_session_dir(session_root, primary_actor)
            if primary_actor
            else session_root
        )
        state_dir = str(actor_root / "state") if primary_actor else ""
    return {
        SESSION_ENV: str(actor_root),
        "GOTTA_SESSION_ID": session_shared_id(session_root),
        CONTENT_ENV: str(dirs.content_dir),
        "GOTTA_SESSION_STATE_DIR": state_dir,
        "GOTTA_SESSION_ACTOR": primary_actor,
    }


def cmd_show(args: argparse.Namespace) -> int:
    dirs = session_dirs_for_read(args)
    require_started_session(dirs)
    return print_session_env(args, show_payload(dirs))



"""GitHub CLI process and payload helpers."""

from __future__ import annotations

import json
import os
import shutil
import subprocess
import sys


def is_interactive() -> bool:
    return sys.stdin.isatty() and sys.stdout.isatty()


def ensure_gh() -> str:
    gh = shutil.which("gh")
    if gh:
        return gh
    raise RuntimeError("missing required dependency: gh")


def ensure_gh_auth(gh: str) -> None:
    status = subprocess.run([gh, "auth", "status"], check=False, capture_output=True)
    if status.returncode == 0:
        return
    if not is_interactive():
        raise RuntimeError("gh is not authenticated. run: gh auth login -w")
    print("GitHub CLI is not authenticated; starting web login...", file=sys.stderr)
    login = subprocess.run([gh, "auth", "login", "-w"], check=False)
    if login.returncode != 0:
        raise RuntimeError("gh auth login failed")


def gh_json(gh: str, args: list[str]) -> bytes:
    env = os.environ.copy()
    env["GH_PAGER"] = "cat"
    proc = subprocess.run([gh, *args], check=False, capture_output=True, env=env)
    if proc.returncode != 0:
        raise RuntimeError(
            proc.stderr.decode("utf-8", errors="replace").strip() or "gh command failed"
        )
    return proc.stdout


def looks_text(data: bytes) -> bool:
    if b"\x00" in data:
        return False
    try:
        data.decode("utf-8")
    except UnicodeDecodeError:
        return False
    return True


def gh_json_object(gh: str, args: list[str]) -> dict[str, object]:
    payload = gh_json_value(gh, args)
    if not isinstance(payload, dict):
        raise RuntimeError("GitHub CLI returned unexpected JSON payload")
    return payload


def gh_json_value(gh: str, args: list[str]) -> object:
    return json.loads(gh_json(gh, args))


def gh_status_payload() -> dict[str, object]:
    gh = shutil.which("gh")
    payload: dict[str, object] = {
        "ghPath": gh or "",
        "ghPresent": bool(gh),
        "authenticated": False,
    }
    if not gh:
        return payload
    proc = subprocess.run(
        [gh, "auth", "status"],
        check=False,
        capture_output=True,
        text=True,
    )
    payload["authenticated"] = proc.returncode == 0
    detail = (proc.stdout or "").strip() or (proc.stderr or "").strip()
    payload["detail"] = detail
    return payload


def default_branch_name(gh: str, *, owner: str, repo: str) -> str:
    payload = gh_json_object(gh, ["api", f"repos/{owner}/{repo}"])
    return str(payload.get("default_branch") or "").strip()



"""Command surface for `gotta session graph`."""

from __future__ import annotations

import argparse
import json

from ..parse import explicit_session_ref, require_started_session, session_dirs_for_read
from .mermaid import render_mermaid
from .payload import graph_payload
from .render import render_graph_text


def cmd_graph(args: argparse.Namespace) -> int:
    dirs = session_dirs_for_read(args)
    require_started_session(dirs)
    payload = graph_payload(
        dirs,
        filter_query=str(getattr(args, "filter", "") or ""),
        session_ref=explicit_session_ref(args),
    )
    if args.output == "json":
        print(json.dumps(payload, indent=2, sort_keys=True))
        return 0
    if args.output == "text":
        print(render_graph_text(payload))
        return 0
    print(render_mermaid(payload))
    return 0



"""Manifest record loading and aggregation."""

from __future__ import annotations

from dataclasses import dataclass, field
import json

from gotta.content.model import ResolvedDirs
from gotta.content.path import content_locator
from gotta.source.visibility import best_visibility_metadata

from ..core import artifact_kind, resolved_visibility_metadata

ManifestEntry = dict[str, object]


def _string(value: object) -> str:
    return str(value or "").strip()


@dataclass(slots=True)
class AggregateState:
    latest: ManifestEntry
    fetch_count: int = 0
    first_fetched_at: str = ""
    last_fetched_at: str = ""
    plugins: set[str] = field(default_factory=set)
    actors: set[str] = field(default_factory=set)
    locators: set[str] = field(default_factory=set)
    artifact_kinds: set[str] = field(default_factory=set)
    visibility: dict[str, object] = field(default_factory=dict)


def manifest_entries(dirs: ResolvedDirs) -> list[ManifestEntry]:
    manifest_path = dirs.content_dir / "manifest.jsonl"
    if not manifest_path.exists():
        return []
    entries: list[ManifestEntry] = []
    for raw_line in manifest_path.read_text(encoding="utf-8").splitlines():
        if not raw_line.strip():
            continue
        payload = json.loads(raw_line)
        if isinstance(payload, dict):
            entries.append(payload)
    return entries


def filter_manifest_entries(
    entries: list[ManifestEntry],
    *,
    plugin: str = "",
    actor: str = "",
    locator: str = "",
) -> list[ManifestEntry]:
    plugin_filter = plugin.strip()
    actor_filter = actor.strip()
    locator_filter = locator.strip()
    filtered: list[ManifestEntry] = []
    for entry in entries:
        if plugin_filter and _string(entry.get("plugin")) != plugin_filter:
            continue
        if actor_filter and _string(entry.get("actor")) != actor_filter:
            continue
        if locator_filter:
            canonical = _string(entry.get("canonical_locator"))
            raw = _string(entry.get("locator"))
            if locator_filter not in canonical and locator_filter not in raw:
                continue
        filtered.append(entry)
    return filtered


def manifest_entry_sort_key(entry: ManifestEntry) -> tuple[str, str, str]:
    return (
        _string(entry.get("fetched_at")),
        _string(entry.get("canonical_locator")) or _string(entry.get("locator")),
        _string(entry.get("checksum")),
    )


def manifest_identity_locator(entry: ManifestEntry) -> str:
    checksum = _string(entry.get("checksum"))
    locator = _string(entry.get("canonical_locator")) or _string(entry.get("locator"))
    if locator:
        return locator
    return content_locator(checksum) if checksum else ""


def aggregate_manifest_entries(entries: list[ManifestEntry]) -> list[ManifestEntry]:
    grouped: dict[tuple[str, str], AggregateState] = {}
    for entry in entries:
        checksum = _string(entry.get("checksum"))
        locator = manifest_identity_locator(entry)
        state = grouped.setdefault((locator, checksum), AggregateState(latest=entry))
        fetched_at = _string(entry.get("fetched_at"))
        if manifest_entry_sort_key(entry) >= manifest_entry_sort_key(state.latest):
            state.latest = entry
        state.fetch_count += 1
        if fetched_at and (
            not state.first_fetched_at or fetched_at < state.first_fetched_at
        ):
            state.first_fetched_at = fetched_at
        if fetched_at and (
            not state.last_fetched_at or fetched_at > state.last_fetched_at
        ):
            state.last_fetched_at = fetched_at
        plugin = _string(entry.get("plugin"))
        actor = _string(entry.get("actor"))
        raw_locator = _string(entry.get("locator"))
        kind = artifact_kind(entry.get("artifact_kind") or entry.get("artifactKind"))
        if plugin:
            state.plugins.add(plugin)
        if actor:
            state.actors.add(actor)
        if raw_locator:
            state.locators.add(raw_locator)
        if locator:
            state.locators.add(locator)
        if kind:
            state.artifact_kinds.add(kind)
        state.visibility = best_visibility_metadata(
            state.visibility,
            resolved_visibility_metadata(
                entry,
                provider=plugin,
                plugin=plugin,
                subcommand=_string(entry.get("subcommand")),
                locator=locator,
            ),
        )

    aggregated: list[ManifestEntry] = []
    for (locator, _checksum), state in grouped.items():
        latest = dict(state.latest)
        latest["canonical_locator"] = (
            _string(latest.get("canonical_locator")) or locator
        )
        latest["fetchCount"] = state.fetch_count
        latest["firstFetchedAt"] = state.first_fetched_at
        latest["lastFetchedAt"] = state.last_fetched_at or _string(
            latest.get("fetched_at")
        )
        latest["fetched_at"] = _string(latest.get("lastFetchedAt"))
        latest["plugins"] = sorted(state.plugins)
        latest["actors"] = sorted(state.actors)
        latest["locators"] = sorted(state.locators)
        latest["artifactKinds"] = sorted(state.artifact_kinds)
        latest.update(best_visibility_metadata(state.visibility))
        aggregated.append(latest)
    return aggregated



"""Requested-disposition payload helpers."""

from __future__ import annotations

from gotta.actor import requested_disposition_label
from gotta.session.registry import _normalize_actor_name
from gotta.session.status.payload.value import ACTOR_TERMINAL_STATUS


def apply_requested_closing(derived_status: str, requested_status: str) -> str:
    if requested_status and derived_status in {
        "starting",
        "active",
        "producing_evidence",
    }:
        return "closing"
    return derived_status


def request_state_payload(
    state: dict[str, object],
    *,
    actor_name: str,
    derived_status: str,
) -> dict[str, object]:
    requested_status = str(state.get("requested_status") or "")
    requested_summary = str(state.get("requested_summary") or "")
    requested_label = requested_disposition_label(state)
    request_note = ""
    if requested_status and derived_status not in ACTOR_TERMINAL_STATUS:
        if derived_status == "stalled":
            request_note = (
                f" Operator already requested `{requested_label}`"
                + (f" ({requested_summary})." if requested_summary else ".")
                + " Because the heartbeat is stale, you can settle now with "
                f"`gotta actor settle {_normalize_actor_name(actor_name)}`."
            )
        else:
            request_note = (
                f" Operator already requested `{requested_label}`"
                + (f" ({requested_summary})." if requested_summary else ".")
                + " That pending disposition will become authoritative automatically "
                "when the actor runtime exits."
            )
    return {
        "requested_status": requested_status,
        "requested_summary": requested_summary,
        "requested_label": requested_label,
        "requested_pending": bool(
            requested_status and derived_status not in ACTOR_TERMINAL_STATUS
        ),
        "request_note": request_note,
    }
